@@ -1,3 +1,4 @@
+import datetime
 from treetime.treeanc import TreeAnc
 from treetime.clock_tree import ClockTree
 from treetime.treetime import TreeTime
@@ -5,8 +6,7 @@ from treetime.treetime import ttconf as treetime_conf
 from treetime.gtr import GTR
 from treetime.treetime import plot_vs_years
 from treetime.treetime import treetime_to_newick
-from treetime import seq_utils
+from treetime.tree_regression import TreeRegression
 from treetime.merger_models import Coalescent
-import datetime
 import treetime.seq_utils as seq_utils
 from treetime.utils import numeric_date
