@@ -1,10 +1,10 @@
-from __future__ import division, print_function
+from __future__ import division, print_function, absolute_import
 import numpy as np
 from scipy.interpolate import interp1d
-from config import BIG_NUMBER, MIN_LOG, MIN_INTEGRATION_PEAK, TINY_NUMBER
 from collections import Iterable
 from copy import deepcopy as make_copy
 from scipy.ndimage import binary_dilation
+from .config import BIG_NUMBER, MIN_LOG, MIN_INTEGRATION_PEAK, TINY_NUMBER
 
 class Distribution(object):
     """
@@ -17,7 +17,7 @@ class Distribution(object):
     """
 
     @staticmethod
-    def calc_fwhm(distribution, is_log=True):
+    def calc_fwhm(distribution, is_neg_log=True):
         """
         Assess the width of the probability distribution. This returns
         full-width-half-max
@@ -25,45 +25,52 @@ class Distribution(object):
 
         if isinstance(distribution, interp1d):
 
-            if is_log:
+            if is_neg_log:
                 ymin = distribution.y.min()
-                prob = np.exp(-(distribution.y-ymin))
+                log_prob = distribution.y-ymin
             else:
-                prob = distribution.y
+                log_prob = -np.log(distribution.y)
+                log_prob -= log_prob.min()
 
             xvals = distribution.x
 
         elif isinstance(distribution, Distribution):
-            # Distribution always stores log-prob
+            # Distribution always stores neg log-prob with the peak value subtracted
             xvals = distribution._func.x
-            prob = distribution.prob_relative(xvals)
+            log_prob = distribution._func.y
         else:
             raise TypeError("Error in computing the FWHM for the distribution. "
                 " The input should be either Distribution or interpolation object");
 
-        x_idxs = binary_dilation(prob >= 0.4*(prob.max() - prob.min())+prob.min(), iterations=1)
-        xs = xvals[x_idxs]
-        if xs.shape[0] < 2:
+        L = xvals.shape[0]
+        # 0.69... is log(2), there is always one value for which this is true since
+        # the minimum is subtracted
+        tmp = np.where(log_prob < 0.693147)[0]
+        x_l, x_u = tmp[0], tmp[-1]
+        if L < 2:
             print ("Not enough points to compute FWHM: returning zero")
             return min(TINY_NUMBER, distribution.xmax - distribution.xmin)
         else:
-            return xs.max() - xs.min()
+            # need to guard against out-of-bounds errors
+            return max(TINY_NUMBER, xvals[min(x_u+1,L-1)] - xvals[max(0,x_l-1)])
 
 
     @classmethod
-    def delta_function(cls, x_pos, weight=1.):
+    def delta_function(cls, x_pos, weight=1., min_width=MIN_INTEGRATION_PEAK):
         """
         Create delta function distribution.
         """
 
-        distribution = cls(x_pos,0.,is_log=True)
+        distribution = cls(x_pos,0.,is_log=True, min_width=min_width)
         distribution.weight  = weight
         return distribution
 
+
     @classmethod
     def shifted_x(cls, dist, delta_x):
         return Distribution(dist.x+delta_x, dist.y, kind=dist.kind)
 
+
     @staticmethod
     def multiply(dists):
         '''
@@ -73,6 +80,7 @@ class Distribution(object):
             raise NotImplementedError("Can only multiply Distribution objects")
 
         n_delta = np.sum([k.is_delta for k in dists])
+        min_width = np.max([k.min_width for k in dists])
         if n_delta>1:
             raise ArithmeticError("Cannot multiply more than one delta functions!")
         elif n_delta==1:
@@ -80,7 +88,7 @@ class Distribution(object):
             delta_dist = dists[delta_dist_ii]
             new_xpos = delta_dist.peak_pos
             new_weight  = np.prod([k.prob(new_xpos) for k in dists if k!=delta_dist_ii]) * delta_dist.weight
-            res = Distribution.delta_function(new_xpos, weight = new_weight)
+            res = Distribution.delta_function(new_xpos, weight = new_weight,min_width=min_width)
         else:
             new_xmin = np.max([k.xmin for k in dists])
             new_xmax = np.min([k.xmax for k in dists])
@@ -95,30 +103,37 @@ class Distribution(object):
                 print ("ERROR in distribution multiplication: Distributions do not overlap")
                 x_vals = [0,1]
                 y_vals = [BIG_NUMBER,BIG_NUMBER]
-                res = Distribution(x_vals, y_vals, is_log=True, kind='linear')
+                res = Distribution(x_vals, y_vals, is_log=True,
+                                   min_width=min_width, kind='linear')
             elif n_points == 1:
                 res = Distribution.delta_function(x_vals[0])
             else:
-                res = Distribution(x_vals[ind], y_vals[ind], is_log=True, kind='linear')
+                res = Distribution(x_vals[ind], y_vals[ind], is_log=True,
+                                   min_width=min_width, kind='linear', assume_sorted=True)
 
         return res
 
 
-    def __init__(self, x, y, is_log=True, kind='linear'):
+    def __init__(self, x, y, is_log=True, min_width = MIN_INTEGRATION_PEAK,
+                 kind='linear', assume_sorted=False):
 
         """
         Create Distribution instance
         """
 
+        self.min_width = min_width
         if isinstance(x, Iterable) and isinstance (y, Iterable):
 
             self._delta = False # NOTE in classmethod this value is set explicitly to True.
-            xvals, yvals = np.array(sorted(zip(x,y))).T
             # first, prepare x, y values
+            if assume_sorted:
+                xvals, yvals = x,y
+            else:
+                xvals, yvals = np.array(sorted(zip(x,y))).T
             if not is_log:
                 yvals = -np.log(yvals)
             # just for safety
-            yvals [np.isnan(yvals)] = BIG_NUMBER
+            yvals[np.isnan(yvals)] = BIG_NUMBER
             # set the properties
             self._kind=kind
             # remember range
@@ -131,7 +146,8 @@ class Distribution(object):
             yvals -= self._peak_val
             self._ymax = yvals.max()
             # store the interpolation object
-            self._func= interp1d(xvals, yvals, kind=kind, fill_value=BIG_NUMBER, bounds_error=False)
+            self._func= interp1d(xvals, yvals, kind=kind, fill_value=BIG_NUMBER,
+                                 bounds_error=False, assume_sorted=True)
             self._fwhm = Distribution.calc_fwhm(self)
 
         elif np.isscalar(x):
@@ -299,7 +315,7 @@ class Distribution(object):
         if n % 2 == 0:
             n += 1
         mult = 1.0/6
-        dpeak = max(5*self.fwhm, MIN_INTEGRATION_PEAK)
+        dpeak = max(10*self.fwhm, self.min_width)
         threshold = np.array([a,self.peak_pos-dpeak, self.peak_pos+dpeak,b])
         threshold = threshold[(threshold>=a)&(threshold<=b)]
         threshold.sort()
