@@ -1,11 +1,9 @@
-#!/usr/local/bin/python
-# -*- coding: utf-8 -*-
-from __future__ import division, print_function
+from __future__ import division, print_function, absolute_import
+from collections import defaultdict
 import numpy as np
-import config as ttconf
-from seq_utils import alphabets, profile_maps, alphabet_synonyms
-from aa_models  import JTT92
-from nuc_models import JC69, K80, F81, HKY85, T92, TN93
+from treetime import config as ttconf
+from .seq_utils import alphabets, profile_maps, alphabet_synonyms
+
 
 class GTR(object):
     """
@@ -20,17 +18,29 @@ class GTR(object):
         ----------
 
          alphabet : str, numpy.array
-            alphabet of the sequence. If string passed, it is understood as
-            alphabet name. In this case, the alphabet and its profile map pulled
-            from seq_utils.
+            Alphabet of the sequence. If a string is passed, it is understood as
+            an alphabet name. In this case, the alphabet and its profile map are pulled
+            from :py:obj:`treetime.seq_utils`.
 
-            If numpy array of characters passed, a new alphabet is constructed,
+            If a numpy array of characters is passed, a new alphabet is constructed,
             and the default profile map is atached to it.
 
+         prof_map : dict
+            Dictionary linking characters in the sequence to the likelihood
+            of observing characters in the alphabet. This is used to
+            implement ambiguous characters like 'N'=[1,1,1,1] which are
+            equally likely to be any of the 4 nucleotides. Standard profile_maps
+            are defined in file seq_utils.py. If None is provided, no ambigous
+            characters are supported.
+
+         logger : callable
+            Custom logging function that should take arguments (msg, level, warn=False),
+            where msg is a string and level an integer to be compared against verbose.
+
         """
         self.debug=False
-        if type(alphabet)==str:
-            if (alphabet not in alphabet_synonyms):
+        if isinstance(alphabet, str):
+            if alphabet not in alphabet_synonyms:
                 raise AttributeError("Unknown alphabet type specified")
             else:
                 tmp_alphabet = alphabet_synonyms[alphabet]
@@ -44,19 +54,43 @@ class GTR(object):
             else:
                 self.profile_map = prof_map
 
+
         if logger is None:
-            def logger(*args,**kwargs):
+            def logger_default(*args,**kwargs):
+                """standard logging function if none provided"""
                 if self.debug:
                     print(*args)
-            self.logger = logger
+            self.logger = logger_default
         else:
             self.logger = logger
-        n_states = len(self.alphabet)
 
+        self.ambiguous = None
+        self.gap_index = None
+        self.n_states = len(self.alphabet)
+        self.assign_gap_and_ambiguous()
+
+        # NEEDED TO BREAK RATE MATRIX DEGENERACY AND FORCE NP TO RETURN REAL ORTHONORMAL EIGENVECTORS
+        # ugly hack, but works and shouldn't affect results
+        tmp_rng_state = np.random.get_state()
+        np.random.seed(12345)
+        self.break_degen = np.random.random(size=(self.n_states, self.n_states))*1e-6
+        np.random.set_state(tmp_rng_state)
+
+        # init all matrices with dummy values
+        self.logger("GTR: init with dummy values!", 3)
+        self.v = None # right eigenvectors
+        self.v_inv = None # left eigenvectors
+        self.eigenvals =None # eigenvalues
+        self.assign_rates()
+
+
+    def assign_gap_and_ambiguous(self):
+        n_states = len(self.alphabet)
         self.logger("GTR: with alphabet: "+str(self.alphabet),1)
         # determine if a character exists that corresponds to no info, i.e. all one profile
         if any([x.sum()==n_states for x in self.profile_map.values()]):
-            self.ambiguous = [c for c,x in self.profile_map.iteritems() if x.sum()==n_states][0]
+            amb_states = [c for c,x in self.profile_map.items() if x.sum()==n_states]
+            self.ambiguous = 'N' if 'N' in amb_states else amb_states[0]
             self.logger("GTR: ambiguous character: "+self.ambiguous,2)
         else:
             self.ambiguous=None
@@ -66,23 +100,15 @@ class GTR(object):
             self.gap_index = list(self.alphabet).index('-')
         except:
             self.logger("GTR: no gap symbol!", 4, warn=True)
-            self.gap_index=-1
-
-
-        # NEEDED TO BREAK RATE MATRIX DEGENERACY AND FORCE NP TO RETURN REAL ORTHONORMAL EIGENVECTORS
-        # ugly hack, but works and shouldn't affect results
-        tmp_rng_state = np.random.get_state()
-        np.random.seed(12345)
-        self.break_degen = np.random.random(size=(n_states, n_states))*1e-6
-        np.random.set_state(tmp_rng_state)
-
-        # init all matrices with dummy values
-        self.logger("GTR: init with dummy values!", 3)
-        self.assign_rates()
+            self.gap_index=None
 
 
     @property
     def Q(self):
+        """function that return the product of the transition matrix
+           and the equilibrium frequencies to obtain the rate matrix
+           of the GTR model
+        """
         return (self.W*self.Pi).T
 
 
@@ -94,24 +120,32 @@ class GTR(object):
         '''
         String representation of the GTR model for pretty printing
         '''
-        eq_freq_str = "Substitution rate (mu): "+str(np.round(self.mu,6))+'\n'
+        multi_site = len(self.Pi.shape)==2
 
-        eq_freq_str += "\nEquilibrium frequencies (pi_i):\n"
-        for a,p in zip(self.alphabet, self.Pi):
-            eq_freq_str+='  '+str(a)+': '+str(np.round(p,4))+'\n'
+        if multi_site:
+            eq_freq_str = "Average substitution rate (mu): "+str(np.round(self.average_rate,6))+'\n'
+        else:
+            eq_freq_str = "Substitution rate (mu): "+str(np.round(self.mu,6))+'\n'
+
+        if not multi_site:
+            eq_freq_str += "\nEquilibrium frequencies (pi_i):\n"
+            for a,p in zip(self.alphabet, self.Pi):
+                eq_freq_str+='  '+str(a)+': '+str(np.round(p,4))+'\n'
 
         W_str = "\nSymmetrized rates from j->i (W_ij):\n"
         W_str+='\t'+'\t'.join(map(str, self.alphabet))+'\n'
         for a,Wi in zip(self.alphabet, self.W):
             W_str+= '  '+str(a)+'\t'+'\t'.join([str(np.round(max(0,p),4)) for p in Wi])+'\n'
 
-        Q_str = "\nActual rates from j->i (Q_ij):\n"
-        Q_str+='\t'+'\t'.join(map(str, self.alphabet))+'\n'
-        for a,Qi in zip(self.alphabet, self.Q):
-            Q_str+= '  '+str(a)+'\t'+'\t'.join([str(np.round(max(0,p),4)) for p in Qi])+'\n'
+        if not multi_site:
+            Q_str = "\nActual rates from j->i (Q_ij):\n"
+            Q_str+='\t'+'\t'.join(map(str, self.alphabet))+'\n'
+            for a,Qi in zip(self.alphabet, self.Q):
+                Q_str+= '  '+str(a)+'\t'+'\t'.join([str(np.round(max(0,p),4)) for p in Qi])+'\n'
 
         return eq_freq_str + W_str + Q_str
 
+
     def assign_rates(self, mu=1.0, pi=None, W=None):
         """
         Overwrite the GTR model given the provided data
@@ -175,12 +209,15 @@ class GTR(object):
          pi : n vector
             Equilibrium frequencies
 
+         **kwargs:
+            Key word arguments to be passed
+
         Keyword Args
         ------------
 
          alphabet : str
             Specify alphabet when applicable. If the alphabet specification is
-            required, but no alphabet specified, the nucleotide will be used as
+            required, but no alphabet is specified, the nucleotide alphabet will be used as
             default.
 
         """
@@ -197,13 +234,9 @@ class GTR(object):
         ----------
 
          model : str
-            Model to create. List of the available models is
-            (see description below)
-
-        Keyword Args
-        ------------
-
-         model arguments
+            Model to create. See list of available models below
+         **kwargs:
+            Key word arguments to be passed to the model
 
 
         **Available models**
@@ -314,6 +347,8 @@ class GTR(object):
                 (kappa1, kappa2) are specified relative to this rate
 
         """
+        from .nuc_models import JC69, K80, F81, HKY85, T92, TN93
+        from .aa_models  import JTT92
 
         if model.lower() in ['jc', 'jc69', 'jukes-cantor', 'jukes-cantor69', 'jukescantor', 'jukescantor69']:
             return JC69(**kwargs)
@@ -336,7 +371,7 @@ class GTR(object):
     @classmethod
     def random(cls, mu=1.0, alphabet='nuc'):
         """
-        Create random GTR model
+        Creates a random GTR model
 
         Parameters
         ----------
@@ -365,7 +400,9 @@ class GTR(object):
         """
         Infer a GTR model by specifying the number of transitions and time spent in each
         character. The basic equation that is being solved is
-            :math:`n_{ij} = pi_i W_{ij} T_j`
+
+        :math:`n_{ij} = pi_i W_{ij} T_j`
+
         where :math:`n_{ij}` are the transitions, :math:`pi_i` are the equilibrium
         state frequencies, :math:`W_{ij}` is the "substitution attempt matrix",
         while :math:`T_i` is the time on the tree spent in character state
@@ -373,14 +410,14 @@ class GTR(object):
         to account for the fact that the root of the tree is in a particular
         state. the modified equation is
 
-            :math:`n_{ij} + pc = pi_i W_{ij} (T_j+pc+root\_state)`
+        :math:`n_{ij} + pc = pi_i W_{ij} (T_j+pc+root\_state)`
 
         Parameters
         ----------
 
          nij : nxn matrix
             The number of times a change in character state is observed
-            between state i and j
+            between state j and i
 
          Ti :n vector
             The time spent in each character state
@@ -393,12 +430,15 @@ class GTR(object):
             Pseudocounts, this determines the lower cutoff on the rate when
             no substitutions are observed
 
+         **kwargs:
+            Key word arguments to be passed
+
         Keyword Args
         ------------
 
          alphabet : str
             Specify alphabet when applicable. If the alphabet specification
-            is required, but no alphabet specified, the nucleotide will be used as default.
+            is required, but no alphabet is specified, the nucleotide alphabet will be used as default.
 
         """
         from scipy import linalg as LA
@@ -426,11 +466,7 @@ class GTR(object):
                                                     + ttconf.TINY_NUMBER + 2*pc_mat)
 
             np.fill_diagonal(W_ij, 0)
-            Wdiag = (((W_ij.T*pi).T).sum(axis=0)+ttconf.TINY_NUMBER)/(pi+ttconf.TINY_NUMBER)
-            np.fill_diagonal(W_ij, Wdiag)
-            Q1 = np.diag(pi).dot(W_ij)
-            scale_factor = np.sum(np.diagonal(Q1*np.diag(pi)))
-            np.fill_diagonal(W_ij, 0)
+            scale_factor = np.einsum('i,ij,j',pi,W_ij,pi)
 
             W_ij = W_ij/scale_factor
             if fixed_pi is None:
@@ -439,14 +475,13 @@ class GTR(object):
             mu = nij.sum()/(ttconf.TINY_NUMBER + np.sum(pi * (W_ij.dot(Ti))))
         if count >= Nit:
             gtr.logger('WARNING: maximum number of iterations has been reached in GTR inference',3, warn=True)
-            np.min(pi.sum(axis=0)), np.max(pi.sum(axis=0))
             if LA.norm(pi_old-pi) > dp:
                 gtr.logger('the iterative scheme has not converged',3,warn=True)
             elif np.abs(1-np.max(pi.sum(axis=0))) > dp:
                 gtr.logger('the iterative scheme has converged, but proper normalization was not reached',3,warn=True)
-        if gtr.gap_index>=0:
+        if gtr.gap_index is not None:
             if pi[gtr.gap_index]<gap_limit:
-              gtr.logger('The model allows for gaps which are estimated to occur at a low fraction of %1.3e'%pi[gtr.gap_index]+
+                gtr.logger('The model allows for gaps which are estimated to occur at a low fraction of %1.3e'%pi[gtr.gap_index]+
                        '\n\t\tthis can potentially result in artificats.'+
                        '\n\t\tgap fraction will be set to %1.4f'%gap_limit,2,warn=True)
             pi[gtr.gap_index] = gap_limit
@@ -493,32 +528,55 @@ class GTR(object):
         self.v = np.real(eigvecs)
         self.v_inv = np.linalg.inv(self.v)
         self.eigenvals = np.real(eigvals)
-        return
 
 
-    def compress_sequence_pair(self, seq_p, seq_ch, pattern_multiplicity=None, ignore_gaps=False):
+    def _eig_sym(self):
+        """
+        Perform eigendecompositon of the rate matrix and stores the left- and right-
+        matrices to convert the sequence profiles to the GTR matrix eigenspace
+        and hence to speed-up the computations.
+        """
+        # eigendecomposition of the rate matrix
+        tmpp = np.sqrt(self.Pi)
+        symQ = self.W*np.outer(tmpp, tmpp)
+        eigvals, eigvecs = np.linalg.eigh(symQ)
+        tmp_v = eigvecs.T*tmpp
+        one_norm = np.sum(np.abs(tmp_v), axis=1)
+        self.v = tmp_v.T/one_norm
+        self.v_inv = (eigvecs*one_norm).T/tmpp
+        self.eigenvals = eigvals
+
+
+    def compress_sequence_pair(self, seq_p, seq_ch, pattern_multiplicity=None,
+                               ignore_gaps=False):
         '''
-        make a compressed representation of a pair of sequences only counting
+        Make a compressed representation of a pair of sequences, only counting
         the number of times a particular pair of states (e.g. (A,T)) is observed
-        the the aligned sequences of parent and child
+        in the aligned sequences of parent and child.
 
         Parameters
         ----------
 
-          seq_p: numpy array
+         seq_p: numpy array
             Parent sequence as numpy array of chars
 
-          seq_ch: numpy array
+         seq_ch: numpy array
             Child sequence as numpy array of chars
 
-          ignore_gap: bool, default False
+         pattern_multiplicity : numpy array
+            If sequences are reduced by combining identical alignment patterns,
+            these multplicities need to be accounted for when counting the number
+            of mutations across a branch. If None, all pattern are assumed to
+            occur exactly once.
+
+         ignore_gap: bool
             Whether or not to include gapped positions of the alignment
             in the multiplicity count
 
         Returns
         -------
           seq_pair : list
-            [(0,1), (2,2), (3,4)] list of parent_child state pairs
+            :code:`[(0,1), (2,2), (3,4)]` list of parent_child state pairs
             as indices in the alphabet
 
           multiplicity : numpy array
@@ -541,24 +599,26 @@ class GTR(object):
                     bs.append(seq==nuc)
 
             for n1,nuc1 in enumerate(self.alphabet):
-                if (n1!=self.gap_index or (not ignore_gaps)):
+                if (self.gap_index is None) or (not ignore_gaps) or (n1!=self.gap_index):
                     for n2,nuc2 in enumerate(self.alphabet):
-                        if (n2!=self.gap_index or (not ignore_gaps)):
+                        if (self.gap_index is None) or (not ignore_gaps) or (n2!=self.gap_index):
                             count = ((bool_seqs_p[n1]&bool_seqs_ch[n2])*pattern_multiplicity).sum()
                             if count: pair_count.append(((n1,n2), count))
         else: # enumerate state pairs of the sequence for large alphabets
-        #FIXME: make this work with pattern_multiplicity
             num_seqs = []
             for seq in [seq_p, seq_ch]: # for each sequence (parent and child) construct a numerical sequence [0,5,3,1,2,3...]
                 tmp = np.ones_like(seq, dtype=int)
                 for ni,nuc in enumerate(self.alphabet):
                     tmp[seq==nuc] = ni  # set each position corresponding to a state to the corresponding index
                 num_seqs.append(tmp)
-            if ignore_gaps:  # if gaps are ingnored skip positions where one or the other sequence is gapped
-                pair_count = Counter([x for x in zip(num_seqs[0], num_seqs[1])
-                                      if (self.gap_index not in x)])
+            pair_count = defaultdict(int)
+            if ignore_gaps:  # if gaps are ignored skip positions where one or the other sequence is gapped
+                for i in range(len(seq_p)):
+                    if self.gap_index!=num_seqs[0][i] and self.gap_index!=num_seqs[1][i]:
+                        pair_count[(num_seqs[0][i],num_seqs[1][i])]+=pattern_multiplicity[i]
             else: # otherwise, just count
-                pair_count = Counter(zip(num_seqs[0], num_seqs[1]))
+                for i in range(len(seq_p)):
+                    pair_count[(num_seqs[0][i],num_seqs[1][i])]+=pattern_multiplicity[i]
             pair_count = pair_count.items()
 
         return (np.array([x[0] for x in pair_count], dtype=int),    # [(child_nuc, parent_nuc),()...]
@@ -570,28 +630,29 @@ class GTR(object):
 ########################################################################
     def prob_t_compressed(self, seq_pair, multiplicity, t, return_log=False):
         '''
-        calculate the probability of observing a sequence pair at a distance t
+        Calculate the probability of observing a sequence pair at a distance t,
+        for compressed sequences
 
         Parameters
         ----------
 
           seq_pair : numpy array
-            np.array([(0,1), (2,2), ()..]) as indicies of
-            pairs of aligned positions. (e.g. 'A'==0, 'C'==1 etc)
-            this only lists all occuring parent-child state pairs, order is irrelevant
+            :code:`np.array([(0,1), (2,2), ()..])` as indicies of
+            pairs of aligned positions. (e.g. 'A'==0, 'C'==1 etc).
+            This only lists all occuring parent-child state pairs, order is irrelevant
 
           multiplicity : numpy array
-            The number of times a parent-child state pair is observed
-            this allows to compress the sequence representation
+            The number of times a parent-child state pair is observed.
+            This allows compression of the sequence representation
 
           t : float
             Length of the branch separating parent and child
 
-          return_log : bool, default False
+          return_log : bool
             Whether or not to exponentiate the result
 
         '''
-        if (t<0):
+        if t<0:
             logP = -ttconf.BIG_NUMBER
         else:
             tmp_eQT = self.expQt(t)
@@ -599,36 +660,41 @@ class GTR(object):
             logQt = np.log(tmp_eQT + ttconf.TINY_NUMBER*(bad_indices))
             logQt[np.isnan(logQt) | np.isinf(logQt) | bad_indices] = -ttconf.BIG_NUMBER
             logP = np.sum(logQt[seq_pair[:,1], seq_pair[:,0]]*multiplicity)
-            if return_log:
-                return logP
-            else:
-                return np.exp(logP)
 
-    def prob_t(self, seq_p, seq_ch, t, pattern_multiplicity = None, return_log=False, ignore_gaps=True):
+        return logP if return_log else np.exp(logP)
+
+
+    def prob_t(self, seq_p, seq_ch, t, pattern_multiplicity = None,
+               return_log=False, ignore_gaps=True):
         """
-        Compute the probability to observe seq_ch after time t starting from seq_p.
+        Compute the probability to observe seq_ch (child sequence) after time t starting from seq_p
+        (parent sequence).
 
         Parameters
         ----------
 
-         profile_p : np.array
-            Parent profile of shape (L, a), where
-            L - length of the sequence, a - alphabet size.
+         seq_p : character array
+            Parent sequence
 
-         profile_ch : np.array
-            Child profile of shape (L, a), where
-            L - length of the sequence, a - alphabet size.
+         seq_c : character array
+            Child sequence
 
          t : double
-            Time (branch len), separating the profiles.
+            Time (branch len) separating the profiles.
 
-         return_log : bool, default False
-            Whether return log-probability.
+         pattern_multiplicity : numpy array
+            If sequences are reduced by combining identical alignment patterns,
+            these multplicities need to be accounted for when counting the number
+            of mutations across a branch. If None, all pattern are assumed to
+            occur exactly once.
+
+         return_log : bool
+            It True, return log-probability.
 
         Returns
         -------
          prob : np.array
-            resulting probability.
+            Resulting probability
 
         """
         seq_pair, multiplicity = self.compress_sequence_pair(seq_p, seq_ch,
@@ -639,6 +705,25 @@ class GTR(object):
     def optimal_t(self, seq_p, seq_ch, pattern_multiplicity=None, ignore_gaps=False):
         '''
         Find the optimal distance between the two sequences
+
+        Parameters
+        ----------
+
+         seq_p : character array
+            Parent sequence
+
+         seq_c : character array
+            Child sequence
+
+         pattern_multiplicity : numpy array
+            If sequences are reduced by combining identical alignment patterns,
+            these multplicities need to be accounted for when counting the number
+            of mutations across a branch. If None, all pattern are assumed to
+            occur exactly once.
+
+         ignore_gaps : bool
+            If True, ignore gaps in distance calculations
+
         '''
         seq_pair, multiplicity = self.compress_sequence_pair(seq_p, seq_ch,
                                                             pattern_multiplicity = pattern_multiplicity,
@@ -646,15 +731,35 @@ class GTR(object):
         return self.optimal_t_compressed(seq_pair, multiplicity)
 
 
-    def optimal_t_compressed(self, seq_pair, multiplicity):
+    def optimal_t_compressed(self, seq_pair, multiplicity, profiles=False, tol=1e-10):
         """
-        Find the optimal distance between the two sequences
+        Find the optimal distance between the two sequences, for compressed sequences
+
+        Parameters
+        ----------
+
+         seq_pair : compressed_sequence_pair
+            Compressed representation of sequences along a branch, either
+            as tuple of state pairs or as tuple of profiles.
+
+         multiplicity : array
+            Number of times each state pair in seq_pair appears (if profile==False)
+
+            Number of times an alignment pattern is observed (if profiles==True)
+
+         profiles : bool, default False
+            The standard branch length optimization assumes fixed sequences at
+            either end of the branch. With profiles==True, optimization is performed
+            while summing over all possible states of the nodes at either end of the
+            branch. Note that the meaning/format of seq_pair and multiplicity
+            depend on the value of profiles.
+
         """
 
         def _neg_prob(t, seq_pair, multiplicity):
             """
-            Probability to observe child given the the parent state, transition
-            matrix and the time of evolution (branch length).
+            Probability to observe a child given the the parent state, transition
+            matrix, and the time of evolution (branch length).
 
             Parameters
             ----------
@@ -678,22 +783,29 @@ class GTR(object):
                 Negative probability of the two given sequences
                 to be separated by the time t.
             """
-            return -1.0*self.prob_t_compressed(seq_pair, multiplicity,t, return_log=True)
+            if profiles:
+                res = -1.0*self.prob_t_profiles(seq_pair, multiplicity,t**2, return_log=True)
+                return res
+            else:
+                return -1.0*self.prob_t_compressed(seq_pair, multiplicity,t**2, return_log=True)
 
         try:
             from scipy.optimize import minimize_scalar
             opt = minimize_scalar(_neg_prob,
-                    bounds=[0,ttconf.MAX_BRANCH_LENGTH],
-                    method='bounded',
-                    args=(seq_pair, multiplicity), options={'xatol':1e-10})
-            new_len = opt["x"]
+                    bounds=[-np.sqrt(ttconf.MAX_BRANCH_LENGTH),np.sqrt(ttconf.MAX_BRANCH_LENGTH)],
+                    args=(seq_pair, multiplicity), tol=tol)
+            new_len = opt["x"]**2
+            if 'success' not in opt:
+                opt['success'] = True
+                self.logger("WARNING: the optimization result does not contain a 'success' flag:"+str(opt),4, warn=True)
         except:
             import scipy
             print('legacy scipy', scipy.__version__)
             from scipy.optimize import fminbound
             new_len = fminbound(_neg_prob,
-                    0,ttconf.MAX_BRANCH_LENGTH,
-                    args=(seq_pair, multiplicity))
+                      -np.sqrt(ttconf.MAX_BRANCH_LENGTH),np.sqrt(ttconf.MAX_BRANCH_LENGTH),
+                       args=(seq_pair, multiplicity))
+            new_len = new_len**2
             opt={'success':True}
 
         if new_len > .9 * ttconf.MAX_BRANCH_LENGTH:
@@ -706,10 +818,54 @@ class GTR(object):
         return new_len
 
 
+    def prob_t_profiles(self, profile_pair, multiplicity, t,
+                        return_log=False, ignore_gaps=True):
+        '''
+        Calculate the probability of observing a node pair at a distance t
+
+        Parameters
+        ----------
+
+          profile_pair: numpy arrays
+            Probability distributions of the nucleotides at either
+            end of the branch. pp[0] = parent, pp[1] = child
+
+          multiplicity : numpy array
+            The number of times an alignment pattern is observed
+
+          t : float
+            Length of the branch separating parent and child
+
+          ignore_gaps: bool
+            If True, ignore mutations to and from gaps in distance calculations
+
+          return_log : bool
+            Whether or not to exponentiate the result
+
+        '''
+        if t<0:
+            logP = -ttconf.BIG_NUMBER
+        else:
+            Qt = self.expQt(t)
+            if len(Qt.shape)==3:
+                res = np.einsum('ai,ija,aj->a', profile_pair[1], Qt, profile_pair[0])
+            else:
+                res = np.einsum('ai,ij,aj->a', profile_pair[1], Qt, profile_pair[0])
+            if ignore_gaps and (self.gap_index is not None): # calculate the probability that neither outgroup/node has a gap
+                non_gap_frac = (1-profile_pair[0][:,self.gap_index])*(1-profile_pair[1][:,self.gap_index])
+                # weigh log LH by the non-gap probability
+                logP = np.sum(multiplicity*np.log(res)*non_gap_frac)
+            else:
+                logP = np.sum(multiplicity*np.log(res))
+
+        return logP if return_log else np.exp(logP)
+
+
     def propagate_profile(self, profile, t, return_log=False):
         """
-        Compute the probability of the sequence state (profile) at time (t+t0),
-        given the sequence state (profile) at time t0.
+        Compute the probability of the sequence state of the parent
+        at time (t+t0, backwards), given the sequence state of the
+        child (profile) at time t0.
 
         Parameters
         ----------
@@ -721,28 +877,62 @@ class GTR(object):
          t : double
             Time to propagate
 
-         return_log: bool, default False
-            Whether to return log-probability
+         return_log: bool
+            If True, return log-probability
 
         Returns
         -------
 
          res : np.array
-            Profile of the sequence after time t.
+            Profile of the sequence after time t in the past.
             Shape = (L, a), where L - sequence length, a - alphabet size.
 
         """
         Qt = self.expQt(t)
         res = profile.dot(Qt)
 
-        if return_log:
-            return np.log(res)
-        else:
-            return res
+        return np.log(res) if return_log else res
 
 
-    def _exp_lt(self, t, mu_prefactor=1.0):
+    def evolve(self, profile, t, return_log=False):
         """
+        Compute the probability of the sequence state of the child
+        at time t later, given the parent profile.
+
+        Parameters
+        ----------
+
+         profile : numpy.array
+            Sequence profile. Shape = (L, a),
+            where L - sequence length, a - alphabet size.
+
+         t : double
+            Time to propagate
+
+         return_log: bool
+            If True, return log-probability
+
+        Returns
+        -------
+
+         res : np.array
+            Profile of the sequence after time t in the future.
+            Shape = (L, a), where L - sequence length, a - alphabet size.
+
+        """
+        Qt = self.expQt(t).T
+        res = profile.dot(Qt)
+        return np.log(res) if return_log else res
+
+
+    def _exp_lt(self, t):
+        """
+        Parameters
+        ----------
+
+         t : float
+            time to propagate
+
         Returns
         --------
 
@@ -752,23 +942,70 @@ class GTR(object):
         """
         return np.exp(self.mu * t * self.eigenvals)
 
+
     def expQt(self, t):
+        '''
+        Parameters
+        ----------
+
+         t : float
+            Time to propagate
+
+        Returns
+        --------
+
+         expQt : numpy.array
+            Matrix exponential of exo(Qt)
+        '''
         eLambdaT = np.diag(self._exp_lt(t)) # vector length = a
-        Qt = self.v.dot(eLambdaT.dot(self.v_inv))   # This is P(nuc1 | given nuc_2)
-        return np.maximum(0,Qt)
+        Qs = self.v.dot(eLambdaT.dot(self.v_inv))   # This is P(nuc1 | given nuc_2)
+        return np.maximum(0,Qs)
+
+
+    def expQs(self, s):
+        eLambdaT = np.diag(self._exp_lt(s**2)) # vector length = a
+        Qs = self.v.dot(eLambdaT.dot(self.v_inv))   # This is P(nuc1 | given nuc_2)
+        return np.maximum(0,Qs)
+
+
+    def expQsds(self, s):
+        '''
+        Returns
+        -------
+        Qtds :  Returns 2 V_{ij} \lambda_j s e^{\lambda_j s**2 } V^{-1}_{jk}
+                This is the derivative of the branch probability with respect to s=\sqrt(t)
+        '''
+        lambda_eLambdaT = np.diag(2.0*self._exp_lt(s**2)*self.eigenvals*s) # vector length = a
+        Qsds = self.v.dot(lambda_eLambdaT.dot(self.v_inv))
+        return Qsds
+
+
+    def expQsdsds(self, s):
+        '''
+        Returns
+        -------
+        Qtdtdt :  Returns V_{ij} \lambda_j^2 e^{\lambda_j s**2} V^{-1}_{jk}
+                This is the second derivative of the branch probability wrt time
+        '''
+        t=s**2
+        elt = self._exp_lt(t)
+        lambda_eLambdaT = np.diag(elt*(4.0*t*self.eigenvals**2 + 2.0*self.eigenvals))
+        Qsdsds = self.v.dot(lambda_eLambdaT.dot(self.v_inv))
+        return Qsdsds
+
 
     def sequence_logLH(self,seq, pattern_multiplicity=None):
         """
-        Returns the log-likelihood of sampling a sequence from equilibrium frequency
-        expects a sequence as numpy array
+        Returns the log-likelihood of sampling a sequence from equilibrium frequency.
+        Expects a sequence as numpy array
 
         Parameters
         ----------
 
          seq : numpy array
-            Compressed sequence as array of chars
+            Compressed sequence as an array of chars
 
-         pattern_multiplicity : numpy_array, None
+         pattern_multiplicity : numpy_array
             The number of times each position in sequence is observed in the
             initial alignment. If None, sequence is assumed to be not compressed
 
@@ -778,6 +1015,8 @@ class GTR(object):
         return np.sum([np.sum((seq==state)*pattern_multiplicity*np.log(self.Pi[si]))
                       for si,state in enumerate(self.alphabet)])
 
+    def average_rate(self):
+        return self.mu*np.einsum('i,ij,j',self.Pi, self.W, self.Pi)
 
     def save_to_npz(self, outfile):
         full_gtr = self.mu * np.dot(self.Pi, self.W)
@@ -790,7 +1029,7 @@ class GTR(object):
     def save_to_json(self, zip):
         d = {
         "full_gtr": self.mu * np.dot(self.Pi, self.W),
-        "Substitution rate" : mu,
+        "Substitution rate" : self.mu,
         "Equilibrium character composition": self.Pi,
         "Flow rate matrix": self.W
         }
